(* This Apply.sml File implements 
 * @ the Rule ModusPonens 
 *     [1] A                                  [1] A 
 *     [2] (A => B)                           [2] (A&C&D ...) => B)
 *      =>  C                                  => C
 *     <apply 1 2> --->             bzw.      <apply 1 2> ---> 
 *     [1] B                                  [2] (C&D ...) => B)
 *     ...                                    ...
 *      => C                                   => C
 *
 * @ Adding of axioms and lemmas to the Hypothesis 
 *   of the goal !
 *      <apply (name)> adds the axiom/lemma (name) 
 *   In case of LEMMA, the lemma has to be shown after the
 *     Main goal proof.
 *   In case of AXIOM not!
 *
 *)


local open Data Goal in
exception nomodusponens
exception noequality


fun appModusPonens_rekursiv (AND(a, r as (AND(b,c)) )) match = 
    if (ETT.equal (a,match)) then r 
    else (AND(a, appModusPonens_rekursiv r match))

  | appModusPonens_rekursiv (AND(a, b)) match = 
	if (ETT.equal(a,match)) then b 
	else if (ETT.equal(b,match)) then a 
	     else raise nomodusponens
  | appModusPonens_rekursiv _ _ = raise nomodusponens

fun appModusPonens (IMP( left as (AND(a,b)), c )) match = IMP( (appModusPonens_rekursiv left match) , c )
  | appModusPonens (IMP(a, b)) match = if (ETT.equal(a,match)) then b else raise nomodusponens
  | appModusPonens _ _ = raise nomodusponens


fun get_subterm (AND(a,b)) i = if i <= 1 then a else get_subterm b (i-1)
  | get_subterm t i = if i = 1 then t else raise Error ("No subterm found !\n")


(* [addLemma]
 * fuegt die Lemmabehauptung als Subgoal hinzu 
 *)
fun addLemma (goal, name) =
    let
	val term = Goal.get_definition(goal, name)
	(* speichere die korrekte Umgebung *)
	val backup_env = Goal.get_env(goal) 
	(* hole die alte Umgebung *)
	val envref = !(Goal.get_envref(goal, name))
	(* setze diese als Hauptumgebung *)
	val goal2 = Goal.set_env(goal, envref)
	(* adde nun Lemma Beweis, hierbei wird Hauptumgebung
	 * dupliziert !! 
	 *)
	val goal3 = Goal.split_goal(goal2, IMP( C "TRUE" ,term ))
	(* schreibe Backup Umgebung zurueck *)
	val goal4 = Goal.set_env(goal3, backup_env)
	(* Aendere diese Definition zu AXIOM *)
	val _ = Goal.change_definitions_kind(goal4, name, AXIOM )
    in
	goal4
    end




(* [process_mp]
 * realisiert Modus Ponens rule 
 *)

fun process_mp (src,dest) goal = 
    let 
	val (left,right) = ( case (Goal.get_first_goal(goal)) of
			    IMP(left,right) => (left,right)
			  | _ => raise Error("Interner Fehler Apply 3 \n" ) )

	(* Bestimme zuerst die src. Voraussetzung *)
	val subsource = get_subterm left src
	(* Dann die Zielvoraussetzung *)
	val subdest   = get_subterm left dest


	val newgoal = ( 
		       let
			   val newprem = appModusPonens subdest subsource
		       in
			   Goal.update_first_goal(goal, IMP(AND(newprem,left), right))
		       end ) handle nomodusponens => goal
    in
	newgoal
    end

(* Zustaendig fuer die Definitionen *)

fun process_defs d goal = 
    let 
	val kind = Goal.get_definitions_kind(goal, d)
	val term = Goal.get_definition(goal, d)
	val newgoal = 	
	    ( case kind of
		  CLASSIC => raise Error ("Standart definitions cannot be applied!\n")

		(* Im Falle eines AXIOMS, fuege es einfach der Praemissen Liste hinzu *)

		| AXIOM   => let 
				 val firstterm = Goal.get_first_goal(goal)
				 val newterm = ( case firstterm of
						(IMP(C "TRUE" ,right)) => (IMP(term,right))
					      | (IMP ( left, right ))  => (IMP(AND(term,left),right))
					      | _ => raise Error("Interner Fehler 1 at Unit Apply")
					        )
				 val goal' = Goal.update_first_goal(goal,newterm)
			     in
				 goal'
			     end

		(* Ein Lemma kann als Voraussetzung hinzugefuegt werden, wenn es im Anschluss
		 * an den Hauptbeweis ebenfalls bewiesen wird 
		 *) 
	     
		| LEMMA   => let 
				 val firstterm = Goal.get_first_goal(goal)
				 val newterm = ( case firstterm of
                                                (IMP(C "TRUE", right)) => (IMP(term,right))
					      |	(IMP(left,right)) => (IMP(AND(term,left),right))
					      | _ => raise Error("Interner Fehler 2 at Unit Apply")
					        )
				 val goal1 = Goal.update_first_goal(goal, newterm)
				 val newgoal = addLemma(goal1, d)
				 
			     in
				 newgoal
			     end
	   ) (* end of case *)
    in 
	     newgoal
    end (* end of process_defs *)


fun main(goal, arg ) = 
    let 
	val new_goal =( case arg of
 	                (DEFINITION d) => process_defs d goal
		      | (MODUSPONENS (src,dest) ) => process_mp (src,dest) goal )

    in
	(Goal.print_goal(new_goal);new_goal)
    end

end (* of local open *)

