(* Bool.sml implementiert die Regel Bool des Natural Deduction
 * calculus; Formeln werden uebersetzt in Primbaeume und so weit
 * wie moeglich vereinfacht.
 * Applikation werden dabei temporaer als Variablen representiert
 * ebenso Teilausdreucke bestehend aus Existenz-oder Allquantoren
 *)

local open Data in

exception primerror
datatype dt = F | T | D of var * dt * dt


(* Implementiert Axiom  < E x. u = u  >  *)
fun redquan (EX((t,s),f))  = if ( ETT.free 0 f ) then (EX((t,s),redquan f))  else redquan (ETT.down 0 f 1)
  | redquan (ALL((t,s),f)) = if ( ETT.free 0 f ) then (ALL((t,s),redquan f)) else redquan (ETT.down 0 f 1)
  | redquan t = t

(* Basisalgorithmen fuer Primbaeume *)
fun red (t as (_, a, b)) = if a=b then a else D(t)

fun var (x: var) = D(x, F, T)

fun neg F = T
  | neg T = F
  | neg (D(x, a, b)) = D(x, neg a, neg b)


fun And (F, _) = F
  | And (T, b) = b
  | And (_, F) = F
  | And (a, T) = a
  | And (a as D(x, a0, a1), b as D(y, b0, b1)) =
    if x=y then
      red(x, And(a0, b0), And(a1, b1))
    else if x<y then red(x, And(a0, b), And(a1, b))
                else red(y, And(a, b0), And(a, b1))

fun or (a,b) = neg(And(neg a, neg b))


(* bcom wandelt Decision Trees back zu pter *)

fun bcom F           v2a = C "FALSE"
  | bcom T           v2a = C "TRUE"
  | bcom (D(x,a,b))  v2a = let val p = (v2a(x)) handle primerror => (V x)   (* Rueckmapping Var to App *)
			     in
				 case (a,b) of 
				     (F,T)   => p
				   | (T,F)   => NOT(p)
				   | (F,b')  => AND(p, bcom b v2a)
				   | (T,b')  => OR(NOT(p), AND(p, bcom b v2a))
				   | (a',T)  => OR(AND(NOT(p), bcom a' v2a), p)
				   | (a',F)  => AND(NOT(p), bcom a' v2a)
				   | (a',b') => OR(AND(NOT(p), bcom a' v2a ), AND(p , bcom b' v2a))
			     end


(* com wandelt pter-Ausdruecke in Primbaeume um *)
(* Mappe Gamma Variablen des Typs Bool ebenfalls zu Boolvars *)

fun com (C "FALSE")               p2v = F
  | com (C "TRUE")                p2v = T
  | com (t as (C s))              p2v = var ( p2v t )
  | com (V(x))                    p2v = var x
  | com (NOT(a))                  p2v = neg(com a p2v)
  | com (AND(a,b))                p2v = And(com a p2v , com b p2v)
  | com (OR(a,b))                 p2v = or (com a p2v , com b p2v)
  | com (IMP(a,b))                p2v = or(neg(com a p2v), com b p2v)
  | com (t as (EQ(a,b)))          p2v = if(ETT.equal(a,b)) then T else var ( p2v t )
  | com (t as (A(_,_)))           p2v = var ( p2v t )
  | com (t as (ALL(_,_)))         p2v = var ( p2v t )
  | com (t as (EX(_,_)))          p2v = var ( p2v t )
  | com (t as GAMMA(TC "bool",_)) p2v = var ( p2v t )  
  | com (L(_,_))                  p2v = raise Error ("Ein Lambda kommt selten allein\n")
  | com t                         p2v = raise Error ("Error while translating pter->primtree\n")
                                                  



(* pterprocess' sammelt Applikationen, Gammavariablen, 
 * Existenz- und Allquantoren in einer Liste rs 
 *)
fun pterprocess' (t as (C("FALSE")) )     rs   = rs
  | pterprocess' (t as (C("TRUE"))  )     rs   = rs
  (* Constanten werden temp. zu Variablen umgetauft *)
  | pterprocess' (t as (C s) )            rs   = (t::rs)
  | pterprocess' (t as (A(a,b)))          rs   = (t::rs)
  | pterprocess' (t as (GAMMA(_)))        rs   = (t::rs)
  | pterprocess' (NOT(a))                 rs   = pterprocess' a rs 
  | pterprocess' (AND(a,b))               rs   = (pterprocess' a rs ) @ (pterprocess' b rs )
  | pterprocess' (OR(a,b))                rs   = (pterprocess' a rs ) @ (pterprocess' b rs )
  | pterprocess' (IMP(a,b))               rs   = (pterprocess' a rs ) @ (pterprocess' b rs )
  | pterprocess' (t as (EQ(_,_)))         rs   = (t::rs)
  | pterprocess' (t as (ALL(_,f)))        rs   = (t::rs)
  | pterprocess' (t as (EX(_,f)))         rs   = (t::rs)
 
  | pterprocess' (L(_,f))                 rs   = rs (* wende BOOL nur auf unterster Stufe an *)
  | pterprocess' _                        rs   = rs

fun pterprocess pter max = let 
                              (* ls beinhaltet alle Subterme des Typs bool, die auf
			         Variablen gemappt werden 
			       *)
			      val ls = (pterprocess' pter nil)
			      val initpter2var = fn pter => raise primerror
			      val initvar2pter = fn i    => raise primerror
				  
			      val (pter2var, var2pter, _ ) = 
				  (      
				   (* Lege Temporaere Mappings fuer die Subterme an *)
				   foldl ( fn (x, (f,g,n) ) => (f x ; (f,g,n)) 
					  handle primerror => ( fn a => if ETT.equal(a,x) then n else f a , 
					               	        fn v => if v=n then x else g v ,
							        n+1 
							      )
					 ) 
				   (initpter2var, initvar2pter, max) ls 
				   )
			  in
			      (* pter2var : pter -> var 
			       * var2pter : var  -> pter 
			       *)
			      (pter2var , var2pter)
			  end
		      


		      
fun main(goal) = let 
		     val firstgoal = Goal.get_first_goal(goal)
		     val n = Goal.nr_freevars(goal)
		     val (pter2var, var2pter) = pterprocess firstgoal n   (* Bekomme Subterm-Mappings *)

		     (* Generiere Primbaum aus firstgoal und mappings *)
		     val tree = com firstgoal pter2var                (* pter -> dec tree *)
		     (* Rueckuebersetzung Primbaum nach Term *)
		     val simplpter' = (bcom tree var2pter)            (* dec tree -> pter *)
		     val simplpter  = case simplpter' of
			                (t as (C "TRUE" )) => t
				      | (t as (C "FALSE")) => t
				      | t                  => IMP (C "TRUE" , t)
			              
		     val newgoal = Goal.update_first_goal(goal, simplpter)
		 in
		     (Goal.print_goal(newgoal);newgoal)

		 end
end (* of local open *)