signature Definition =
sig
    type definition
    exception no_entry

    (* Gibt einen initialen Definitionen Container zurueck *)
    val empty : unit -> definition

    (* Adds a new definition *)
    val insert_def : definition -> (string * Data.pter * Data.kind_of_def) -> Env.env ref -> definition

    (* Returns the Env ref for a saved definition with key *)
    val get_envref : definition -> string -> Env.env ref

    (* Gets a term/definition by an key *)
    val get_term : definition -> string -> Data.pter

    (* Returns the Kind of a saved Definition : CLASSIC | LEMMA | AXIOM *)
    val get_kind : definition -> string -> Data.kind_of_def

    (* Changes the kind of a definition i.e a lemma into an axiom, after having used it *)
    val change_kind : definition -> (string * Data.kind_of_def) -> unit

    (* Returns the number of known definitions *)
    val nr_def : definition -> int

    (* Checks, if there are definitions inside the term pter *)
    val check_def : definition -> (int -> string) -> Data.pter -> bool

    (* Substituate the definitions in the term pter *)
    val translate_def : definition -> (int->string)-> Data.pter -> Data.pter

end