(* Elim.sml implementiert alle Eliminierungs-Regeln des Natural Deduction
 * calculus:
 *
 *  @  (E x . A[x]) => B   -->  A[x] => B                 with x as a new parameter
 *  @  (\/ x. A[x]  => B   -->  A[? i] => B               (? i) is the i-th gamma variable
 *  @  A + B => C          -->  (1) A => B   (2) A => C   splits the goal in two subgoals
 *  @  (A & B) & C  => D   -->  B & ( A & C )             smoothes the and-tree
 *  @  ~A => B             -->  ~A => A                   if ~A is valid, then it's enough to show A
 *)



local open Data Goal in

fun main(goal,modus) = 
    let 
	val pter = Goal.get_first_goal(goal)
	val newgoal = elim_left pter goal modus
    in (Goal.print_goal(newgoal); newgoal)
    end


and elim_left (IMP(oldleft as EX((t,i),m), right)) goal _ =                              (* EXISTIERT LEFT *)
    let 
	val newname    = Goal.freeVarIntro(goal,i) (* holt den String namen der geb. Var i *)
	val tempgoal = Goal.update_ftypeenv(goal, newname , t) (* hier werden ebenfalls die geb. Var ersetzt *)
	val freemaps2i  = Goal.freemaps2i(tempgoal)
	val tempgoal' = Goal.add_delta(tempgoal, freemaps2i newname)
	val newleft  = ETT.down 0 (ETT.sub m 0 (V ((freemaps2i newname) + 1))) 1 
	val tempgoal''= Goal.update_first_goal(tempgoal',IMP( newleft, right ))
	val newgoal = Goal.kaper_check(tempgoal'')
    in
	newgoal
    end
  | elim_left (IMP(oldleft as AND(EX((t,i),m),B), right)) goal _ =                       (* EXISTIERT LEFT IM AND CASE*)
    let 
	val newname    = Goal.freeVarIntro(goal,i) (* holt den String namen der geb. Var i *)
	val tempgoal   = Goal.update_ftypeenv(goal, newname , t) (* hier werden ebenfalls die geb. Var ersetzt *)
	val freemaps2i  = Goal.freemaps2i(tempgoal)
	val tempgoal' = Goal.add_delta(tempgoal, freemaps2i newname)
	val newleft' = ETT.down 0 (ETT.sub m 0 (V ((freemaps2i newname) + 1))) 1
	val newleft  = AND(newleft', B)
	val tempgoal''= Goal.update_first_goal(tempgoal',IMP(newleft,right ))
	val newgoal = Goal.kaper_check(tempgoal'')
    in
	newgoal
    end

  | elim_left (IMP(oldleft as ALL((t,i), m), right)) goal modus =                            (* ALL LEFT *)
    let
	val nr_next_gamma = Goal.next_gamma_nr(goal)
	val instance = GAMMA (t, nr_next_gamma)
	val newgoal' = Goal.add_gamma(goal)
	val newleft' = ETT.down 0 (ETT.sub m 0 (instance)) 1
	val newleft = ( case modus of
		       STANDART => newleft'
		     | MORE => AND(newleft', oldleft) )
	val newgoal = Goal.update_first_goal(newgoal',IMP( newleft , right))
    in
	newgoal
    end

  | elim_left (IMP(oldleft as AND(ALL((t,i), m), B), right)) goal modus =                     (* ALL LEFT IM AND CASE *)
    let
	val nr_next_gamma = Goal.next_gamma_nr(goal)
	val instance = GAMMA (t, nr_next_gamma)
	val newgoal' = Goal.add_gamma(goal)
	val newleft' = ETT.down 0 (ETT.sub m 0 (instance)) 1
	val newleft  = ( case modus of 
			STANDART => AND( newleft', B )
		      | MORE => AND(newleft', oldleft) )
	val newgoal = Goal.update_first_goal(newgoal',IMP(newleft , right))
    in
	newgoal
    end

  | elim_left (IMP(OR(a,b), right)) goal _ =                              (*  A + B => C  wird zu  A => C und B => C *)
    let
	val newgoal1 = Goal.update_first_goal(goal, IMP(a,right))
	val newgoal2 = Goal.split_goal(newgoal1, IMP(b,right))
    in
	newgoal2
    end

  | elim_left (IMP(AND(OR(a,b),c), right)) goal _ =     
    let
	val newgoal1 = Goal.update_first_goal(goal, IMP(AND(a,c),right))
	val newgoal2 = Goal.split_goal(newgoal1, IMP(AND(b,c),right))
    in
	newgoal2
    end

  | elim_left (IMP(IMP(a,b), right)) goal _ =                     (* ( A => B ) => C wird zu B => C und (~C) => A *)
    let
	val newgoal1 = Goal.update_first_goal(goal, IMP(b, right))
	val newgoal2 = Goal.split_goal(newgoal1, IMP(NOT(right), a))
    in 
	newgoal2
    end
 
  | elim_left (IMP(AND(IMP(a,b),c), right)) goal _ = 
    let
	val newgoal1 = Goal.update_first_goal(goal, IMP(AND(b,c), right))
	val newgoal2 = Goal.split_goal(newgoal1, IMP(AND(NOT(right),c), a))
    in 
	newgoal2
    end





  | elim_left (IMP(left as (NOT(a)), right )) goal _ =             (* Wenn wir ~A haben, genuegt es A zu zeigen *)
     let 
	 val newright = a
	 val newgoal = Goal.update_first_goal(goal,IMP( left , newright ))
     in 
	 newgoal
     end

  | elim_left (IMP(left as (AND(NOT(a),b)), right )) goal _ =
     let
	 val newright = a
	 val newgoal = Goal.update_first_goal(goal, IMP(left,newright))
     in
	 newgoal
     end
  

  | elim_left (IMP(AND(AND(a,b), c), right)) goal _ =        (* glaettet den AND Tree (A * B) * C => D *)
     let                                                   (* wird zu                 B * A * C => D *)
	val newleft = AND(b,AND(a,c))
	val newgoal = Goal.update_first_goal(goal, IMP(newleft,right))
     in 
	 newgoal
     end

	 
  | elim_left (IMP(t as (EQ(a,b)), right )) goal _ =
     let
	 (* 
	  * Rufe Prozedur Equality.main auf
	  * gib den ersetzten Subterm zurueck
	  *)
	 val (newleft, newgoal) = Equality.main(t, goal)  
     in
	 Goal.update_first_goal(newgoal, IMP(newleft, right))
     end
	 
  | elim_left (IMP(AND(t as (EQ(a,b)), B), right )) goal _ =
     let
	 (* 
	  * Rufe Prozedur Equality.main auf
	  * gib den ersetzten Subterm zurueck
	  *)
	 val (newleft, newgoal) = Equality.main(t, goal)  
     in
	 Goal.update_first_goal (newgoal, IMP(AND(newleft, B), right ))
     end

     
(* TRANSLATE DEFINITION CHECK 
 *
 * Handelt es sich bei der obersten Praemisse um eine Definition des users,
 * wird versucht, den Ausdruck der Definition einzusetzen: 
 * A[D] => B  and A contains a Definition (D=C) --> A[D:=C] => B
 *)

  | elim_left (IMP(left, right )) goal modus =
     let
	 val (newleft, manipulated) = 
	     case left of
		 AND(a,b) => if (Goal.check_def(goal, a)) then (AND(Goal.translate_def(goal,a),b) , true)
			     else (left, false)
	       | a => if (Goal.check_def(goal, a)) then (Goal.translate_def(goal,a), true)
		      else (left, false)
	 val newgoal = if manipulated then Goal.update_first_goal(goal, IMP(newleft, right)) else goal
     in 
     (* Damit kein Zyklus entsteht *)
	 if manipulated then 
	     let val newfirst = Goal.get_first_goal(newgoal)
	     in elim_left newfirst newgoal modus
	     end 
	 else newgoal   

     end


  | elim_left _ goal _ = goal
	     
end (* of local open *)