(* Struktur Goal.sml implementiert Signatur Goal.sig *)
(* Erklaerung siehe Goal.sig *)

structure Goal :> Goal =
struct
  open Data
  type goal =  Env.env list * pter list * Gamma.gamma * Definition.definition
  (*    goal ist ein 4-Tupel aus :
        Typumgebung * Praeterme * Gamma Variablen
        * Definition Hash Table
    
   *    [env1,env2,...] gehoert zu [pter1,pter2,...]    
   *    
   *)

(* --------------------------------------------------------------------*)
(* |                   BEGIN Hilfsprozeduren                          |*)
(* --------------------------------------------------------------------*)


  (* Signalisiert Main.sml, dass Proof Done ! *)      
  fun done_goal() =
      let
	  val done_gamma = Gamma.empty()     (* Gamma Graph inital leer *)
	  val done_def   = Definition.empty()  (* lege leere Definitionen table an *)
	  val done_pter  = (C "PROOF DONE")
      in
	 ( nil , done_pter::nil , done_gamma , done_def )  
      end
     
  (* Erste Uebersetzung der geparsten Ausdruecke in Uebergangsrepresentation 
   * sexpr -> pter' 
  
  fun sexpr2pter' (AppS(a,b))     map = A'(sexpr2pter' a map, sexpr2pter' b map)
    | sexpr2pter' (CstS(s) )      map = C'(s)
    (*  Ist eine Variable nicht in gesmaps2i ==> Konstante oder unbekannt *)
    (*  Im Falle unbekannt, merkt der Typchecker dies !! *)
    | sexpr2pter' (VarS(s) )      map = ( ( V'(map s) ) handle Error _ => (C' (s)) )
    | sexpr2pter' (LamS((s,t),e)) map = L'((map s,t), sexpr2pter' e map)
    | sexpr2pter' (ExS((s,t),e))  map = EX'((map s,t), sexpr2pter' e map)
    | sexpr2pter' (AllS((s,t),e)) map = ALL'((map s,t), sexpr2pter' e map)
    | sexpr2pter' (AndS(a,b))     map = AND'(sexpr2pter' a map, sexpr2pter' b map)
    | sexpr2pter' (OrS(a,b))      map = OR'(sexpr2pter' a map, sexpr2pter' b map)
    | sexpr2pter' (ImpS(a,b))     map = IMP'(sexpr2pter' a map, sexpr2pter' b map)
    | sexpr2pter' (EqS(a,b))      map = EQ' (sexpr2pter' a map, sexpr2pter' b map)
    | sexpr2pter' (NotS a)        map = NOT'(sexpr2pter' a map)
  *)    
  (* Prueft, ob Elemente in xs bereits in map deklariert wurden *)
(*  fun kaper_free xs map = foldl ( fn ( x, b) => (map x; false) handle Error _ => b ) true xs *)
 

(* Returns  Bound Variable List *)
  fun allbndvar' (AppS( a, b ))  bs = (allbndvar' a bs) @ (allbndvar' b bs)
    | allbndvar' (CstS _)        bs = bs
    | allbndvar' (VarS _)        bs = bs	  
    | allbndvar' (LamS((s,t),e)) bs = (allbndvar' e (s::bs))
    | allbndvar' (ExS((s,t),e))  bs = (allbndvar' e (s::bs))
    | allbndvar' (AllS((s,t),e)) bs = (allbndvar' e (s::bs))
    | allbndvar' (AndS(a,b))     bs = (allbndvar' a bs) @ (allbndvar' b bs)
    | allbndvar' (OrS(a,b))      bs = (allbndvar' a bs) @ (allbndvar' b bs)
    | allbndvar' (ImpS(a,b))     bs = (allbndvar' a bs) @ (allbndvar' b bs)
    | allbndvar' (EqS(a,b))      bs = (allbndvar' a bs) @ (allbndvar' b bs)
    | allbndvar' (NotS(a))       bs = allbndvar' a bs
  fun allbndvar sexpr = allbndvar' sexpr nil
      
  (* String Ordnung *)
  val str_order = fn (x:string,y:string) => if x < y then LESS else if x > y then GREATER else EQUAL

  (* remove_dbl loescht das Doppelauftreten der Variablen in der Liste *)
  fun remove_dbl xs = foldr ( fn (x,nil) => [x] | (x,ys) => if x = hd(ys) then ys else x::ys ) 
      nil (Listsort.sort str_order xs)


  fun sexpr2pter (AppS(a,b))     trans map typmap = A(sexpr2pter a trans map typmap, sexpr2pter b trans map typmap)
    | sexpr2pter (CstS(s) )      trans map typmap = C(s)
    (*  Ist eine Variable nicht in der map ==> Konstante oder unbekannt *)
    (*  Im Falle unbekannt, merkt der Typchecker dies !! *)
    | sexpr2pter (VarS(s) )      trans map typmap = ((V (map s)) handle Error _ => (C(s)))
    | sexpr2pter (LamS((s,t),e)) trans map typmap = 
      let 
	  val newtype = if (t=TC("not_specified")) 
			then ((typmap s) handle Error _ => TC ("bool"))
			else t  
      in 
	  L((newtype,(trans s)), sexpr2pter e trans (fn t => if t=s then 0 else (map t) + 1) typmap )
      end
    | sexpr2pter (ExS((s,t),e))  trans map typmap = 
      let 
	  val newtype = if (t=TC("not_specified")) 
			then ((typmap s) handle Error _ => TC ("bool"))
			else t  
      in 
	  EX((newtype,(trans s)), sexpr2pter e trans (fn t => if t=s then 0 else (map t) + 1) typmap )
      end
    | sexpr2pter (AllS((s,t),e)) trans map typmap = 
      let 
	  val newtype = if (t=TC("not_specified")) 
			then ((typmap s) handle Error _ => TC ("bool"))
			else t  
      in 
	  ALL((newtype,(trans s)), sexpr2pter e trans (fn t => if t=s then 0 else (map t) + 1) typmap )
      end
    | sexpr2pter (AndS(a,b))     trans map typmap = AND(sexpr2pter a trans map typmap, sexpr2pter b trans map typmap )
    | sexpr2pter (OrS(a,b))      trans map typmap = OR (sexpr2pter a trans map typmap, sexpr2pter b trans map typmap )
    | sexpr2pter (ImpS(a,b))     trans map typmap = IMP(sexpr2pter a trans map typmap, sexpr2pter b trans map typmap )
    | sexpr2pter (EqS(a,b))      trans map typmap = EQ (sexpr2pter a trans map typmap, sexpr2pter b trans map typmap )
    | sexpr2pter (NotS a)        trans map typmap = NOT(sexpr2pter a trans map typmap)




     
(* --------------------------------------------------------------------*)
(* |                   ENDE Hilfsprozeduren                           |*)
(* --------------------------------------------------------------------*)
      
      
      

(* --------------------------------------------------------------------*)
(* |        BEGIN Methoden zur Bearbeitung des First Goal             |*)
(* --------------------------------------------------------------------*)



(* [init_goal]
 * legt initiales Goal an 
 *)
  fun init_goal() = 
      let 
	  val init_gamma = Gamma.empty()     (* Gamma Graph inital leer *)
	  val init_env = Env.empty()         (* lege leere Typ/Var.umgebung an *)
	  val init_def = Definition.empty()  (* lege leere Definitionen table an *)
		 
      in ( [init_env] , nil, init_gamma , init_def )    (* Goalliste list ist initial NIL *)
      end

 
 
  (* [init_goal_env]
   * legt initiales Goal mithilfe eines alten Goals an
   *)
  fun init_goal_env(oldgoal) = 
      let   
	  val (old_envs, _ , old_gamma, olddef)  = oldgoal
      in ( old_envs, nil, old_gamma, olddef)
      end
  
 (* [ first_goal ]
  *  - scanne geparsten Ausdruck und update entsprechend gebundene Variablen
  *  - Umwandlung von sexpr -> pter
  *  - Stelle pter A im Implikativen Format dar : TRUE => A
  *)

  fun first_goal(oldgoal, sexpr) = 
      let 
	  val (type_envs, pters, gamma, def) = oldgoal
	  (* type_envs liste invariant nicht leer *)
	  val type_env = hd(type_envs) handle empty => raise Error("Environmentlist empty!")
	  
	  (* scanne nach gebundenen Variablen *)
	  val bndxs = remove_dbl( allbndvar sexpr )

	  (* !!! Mache die neue Boundvar Liste der Umgebung bekannt !!! *)
     	  val newenv = Env.update_bndmaplist(type_env, bndxs)
	      
	  (* Mapping von Freelist@Bndlist temporaer *)
	  val freemaps2i = Env.freemaps2i(newenv) 
	  val bndmaps2i  = Env.bndmaps2i(newenv)
	  val freetypmap = fn s => Env.typf(newenv, s)

	  (* Beginne Uebersetzung in DeBrujn Darstellung *)
	  val newpter'  = sexpr2pter sexpr bndmaps2i freemaps2i freetypmap
	  (* Intern liegen alle Terme /Formeln im Deduktionsformat vor *)
	  val newpter   = IMP ( C "TRUE" , newpter' )  
	 (* Debug
	  val _ = print ("\n\n" ^ ( Visual.pter2string newpter)  ^ "\n\n")
	  *)
      in
	  (* Typecheck wird spaeter durchgefuehrt *) 
	  ( newenv::tl(type_envs), pters@[newpter], gamma, def )  
      end
  


  (* [ get_first_goal ]
   * liefert den zum ersten Goal gehoerigen Ausdruck 
   *)
  fun get_first_goal(oldgoal) = 
      let 
	  val (_, pters,_,_) = oldgoal
      in 
	  hd(pters) handle Empty => raise Error "no avtive goals\n"
      end
  


 (* [ update_first_goal ]
  * ersetzt das erste goal durch einen Ausdruck pter 
  *)
  fun update_first_goal(oldgoal, pter) = 
      let 
	  val (type_envs, pters, gamma, def ) = oldgoal
	  val type_env = hd(type_envs) handle empty => raise Error("Environmentlist empty!")
	      
      in 
	  if (null(pters)) then raise Error "First goal doesn't exist, so it can't be updated \n"
	  else 
	      (* Stimmen die Typen des FirstGoal und des 
	       * Ausdrucks pter ueberein, darf ersetzt werden 
	       *)
	      if ( Typecheck.tycheck pter ( (fn c => Env.typc (type_env, c)),
		                            (fn i => Env.typi (type_env, i)),
		                            (Env.typmaps2b(type_env))         ) )
		  then ( case pter of
			(C "TRUE") => if (null(tl(pters)) andalso null(tl(type_envs))) 
					  then ( print ("> PROOF DONE\n"); done_goal() )
				      else (tl(type_envs), tl(pters), gamma, def)
					  
		      | (C "FALSE") => (print ("> THEOREM NOT TRUE try it again \n"); init_goal() )
		      | p => (type_envs, (p::(tl(pters))), gamma ,def)
					  )
	      else raise Error ("Formula is illtyped, must have type bool")
		 
      end 

(* [remove_first_goal]
 * loescht das erste Goal und entsprechende Umgebung
 *)

  fun remove_first_goal ( goal ) =
      let
  	  val (type_envs, pters, gamma, def) = goal
      in
	  if (null(type_envs) orelse null(pters) ) then 
	      raise Error ("There is no first goal removeble!\n")
	  else
	      (tl(type_envs), tl(pters), gamma, def)
      end
		      


(* [typecheck]
 * prueft, ob der Typ des ersten Goals Bool ist
 *)
  fun typecheck(goal) = 
      let
	  val (type_envs, pters,_,_) = goal
	  val type_env = hd(type_envs)
      in
	  if (null(pters)) 
	      then raise Error "First goal doesn't exists, so it cannot be typechecked\n"
	  else 
	      ( Typecheck.tycheck (hd(pters)) ( (fn c => Env.typc (type_env, c)),
					        (fn i => Env.typi (type_env, i)), 
						(Env.typmaps2b(type_env))        ) )
      end
  
(* --------------------------------------------------------------------*)
(* |        ENDE Methoden zur Bearbeitung des First Goal              |*)
(* --------------------------------------------------------------------*)


(* --------------------------------------------------------------------*)
(* |        BEGIN Methoden zur Erweiterung der Goalliste              |*)
(* --------------------------------------------------------------------*)

  (* [ split_goal ] 
   * erweitere die Liste der Goals, indem der Ausdruck pter
   * als weiteres Subgoal hinzugefuegt wird.
   *)
  
  fun split_goal(oldgoal, new_pter) = 
      let 
	  val (type_envs, pters, gamma, def) = oldgoal
	  val type_env = hd(type_envs)
	  val pter     = hd(pters)

	  (* fuege das subgoal nach dem aktuellen ein
	   * und dupliziere die Umgebung !!!
	   *)
	  val new_type_envs = type_env::type_env::(tl(type_envs))
	  val new_pters     = pter::new_pter::(tl(pters))

      in 
	  (* New_Pter muss den Typ Bool haben *)
	  if ( Typecheck.tycheck new_pter ( (fn c => Env.typc (type_env, c)),
	                                    (fn i => Env.typi (type_env, i)), 
	                                    (Env.typmaps2b(type_env))        ) )
	    (*   then (type_envs@[type_env], pters@[pter], gamma, def)  *)
	      then ( new_type_envs, new_pters, gamma, def )
	  (* Dupliziere Umgebungen *)
	  else (type_envs, pters, gamma, def)
      end


  (* [ add_goal ] 
   * erweitere die Liste der Goals, indem der erste Ausdruck des addgoals
   * zusammen mit seiner Umgebung als weiteres Subgoal hinzugefuegt wird.
   *)
  
  fun add_goal(oldgoal, addgoal) = 
      let 
	  val (type_envs, pters, gamma, def) = oldgoal
	  val (add_type_envs, add_pters, _ ,_ ) = addgoal
	  val add_type_env = hd(add_type_envs)
	  val add_pter    = hd(add_pters)
	     
	  (* fuege das subgoal am Ende ein  *)
	  val new_type_envs = type_envs @ [add_type_env]
	  val new_pters     = pters @ [add_pter]

      in 
	  ( new_type_envs, new_pters, gamma, def )
      end



  
 (* [update_ctypeenv]
   * Konstante s mit Typ t, wird der Konstanten/Typumgebung hinzugefuegt
   * hierzu werden Methoden der Struktur Env aufgerufen
   *)

  fun update_ctypeenv(oldgoal, s, t) = 
      let 
	  val (old_envs, pters, gamma,def) = oldgoal
	  val old_env = hd(old_envs)

	  val newenv' = Env.update_ctypeenv(old_env,s,t)
	  val newenv  = Env.update_constmaplist(newenv',[s])

      in
	  ( newenv::tl(old_envs), pters, gamma, def)
      end
 
  (* [update_ftypeenv]
   * Variable s mit Typ t, wird der Variablen/Typumgebung hinzugefuegt
   * hierzu werden Methoden der Struktur Env aufgerufen
   *)

   fun update_ftypeenv (oldgoal, s, t) = 
      let 
	  val (old_envs, pters, gamma,def) = oldgoal
	  val old_env = hd(old_envs)

	  val newenv' = Env.update_ftypeenv(old_env,s,t)
	  val newenv  = Env.update_freemaplist(newenv',[s])

      in
	  ( newenv::tl(old_envs), pters, gamma, def)
      end


 (* [update_bndvar]
   * Gebundene Variable mit Namen s wird der Umgebung hinzugefuegt
   * Diese Methode wird lediglich zur Uebersetzung von Gleichheitstermen
   * gebraucht
   *)
  fun update_bndvar (oldgoal, s) =
      let 
	  val (old_envs, pters, gamma, def) = oldgoal
	  val old_env = hd(old_envs)
	  val new_env = Env.update_bndmaplist(old_env, [s])
      in
	  ( new_env::tl(old_envs) , pters , gamma, def )
      end

  (* [new_type]
   * Typ s wird der Typliste hinzugefuegt
   * hierzu wird die Methode Env.new_type der Struktur Env aufgerufen
   *)

  fun new_type(oldgoal, s) = 
      let 
	  val (old_envs, pters, gamma, def) = oldgoal
	  val old_env = hd(old_envs)
	  val new_env = Env.new_type(old_env,s)
      in
	  ( new_env::tl(old_envs) , pters , gamma, def )
      end

 


 (* [kaper_check]
   * Checkt, ob freie Variablen durch Instanziierungen in gebundenen Ausdruecken 
   * gekapert wurden.
   * Checkt lediglich ERSTES Goal  !! Achtung !!
   * Die Umbenennung kann mehrmals erfolgen, da es zu einem Dominoeffekt kommen
   * kann -> aber terminiert, da Freie Variablen endlich !!
   *)

  fun kaper_check(oldgoal) = 
      let
	  val (old_envs, pters, gamma, def) = oldgoal
	  val old_env = hd(old_envs) 
	  val pter = hd(pters)

	  val firstbndmapi2s   = Env.bndmapi2s(old_env)
	  val freemapi2s       = Env.freemapi2s(old_env)
	  val constmapi2s      = Env.constmapi2s(old_env)
	  val firstmaps = (firstbndmapi2s, freemapi2s, constmapi2s)
	      
	  fun loopUempty env nil = env
	    | loopUempty env cs  = 
	      let val new_env   = Env.renameBndVars(env, cs)
		  val bndmapi2s = Env.bndmapi2s(new_env)
		  val maps      = (bndmapi2s, freemapi2s, constmapi2s)
		  val newcs     = ETT.getCriticalBnds pter maps
	      in loopUempty new_env newcs 
	      end
	  
	  val crit_bnds = ETT.getCriticalBnds pter firstmaps
	 (* DEBUG val _ = foldl( fn (i,_) => print (Int.toString(i) ^ " ") ) () crit_bnds *)
	  val new_env =  loopUempty old_env crit_bnds
	      
      in
	  (new_env::tl(old_envs), pters, gamma, def)

      end

(* [freeVarIntro]
 * calls the Env.freeVarIntro
 * more information there
 *)

  fun freeVarIntro(oldgoal, i) =
      let
	  val (old_envs, pters, gamma, def) = oldgoal
	  val oldenv = hd(old_envs) 
      in
	  Env.freeVarIntro(oldenv, i)
      end


(* --------------------------------------------------------------------*)
(* |        ENDE Methoden zur Erweiterung der Goalliste               |*)
(* --------------------------------------------------------------------*)

(* --------------------------------------------------------------------*)
(* |                        BEGIN Gamma Methoden                      |*)
(* --------------------------------------------------------------------*)


(* [next_gamme_nr]
 * liefert die naechste freie Gamma Nr 
 *)

  fun next_gamma_nr(oldgoal) = 
      let
	  val (_,_,gamma,_) = oldgoal
      in
	  (* Zaehle ab 0 *)
	  Gamma.nr_gammavars(gamma) 
      end


  (* [add_delta] 
   * Eine neue Delta Variable wird dem Graphen hinzugefuegt
   * Umgebung hinzugefuegt
   *)
  fun add_delta(oldgoal, i) = 
      let
	  val (envs,pters,gamma,def) = oldgoal
	  val newgamma = Gamma.add_delta(gamma, i)
      in
	 (envs, pters, newgamma, def)
      end
  


 (* [add_gamma] 
  * Eine neue Gamma Variable wird dem Graphen hinzugefuegt
  * Umgebung hinzugefuegt
  *)
  fun add_gamma(oldgoal) = 
      let
	  val (envs,pters,gamma,def) = oldgoal
	  val nextnr = Gamma.nr_gammavars(gamma)
	  val newgamma = Gamma.add_gamma(gamma, nextnr)
      in
	  (envs, pters, newgamma, def)
      end
  

  (* [substitute_gamma] 
   * gibt true zurueck, wenn die Ersetzung ok ist
   * false, wenn es nicht moeglich zu ersetzen
   *)
  fun substitute_gamma oldgoal ( i, vars) =
      let
	  val (_,_,gamma,_) = oldgoal
      in
	  Gamma.substitute gamma (i,vars)
      end

  (* [merge_envs]
   * mergt nur die Umgebungen des oldgoals und des instgoals, so dass
   * - BndVarMap des Instgoals uebernommen werden
   * - FreeVarMap, Typelist und Typemap des oldgoals
   *)

  fun merge_envs(oldgoal, instgoal) = 
      let
	  val (old_envs, pters, old_gamma, def) = oldgoal
	  val old_env = hd(old_envs)
	  val (inst_envs, _, _ ,_) = instgoal	  
	  val inst_env = hd(inst_envs)
	  val merge_env = Env.back_instance_env(old_env,inst_env)
      in
	  (merge_env::tl(old_envs), pters, old_gamma, def)
      end


(* --------------------------------------------------------------------*)
(* |                        ENDE Gamma Methoden                       |*)
(* --------------------------------------------------------------------*)
 

(* --------------------------------------------------------------------*)
(* |                     BEGIN Definitionen Methoden                  |*)
(* --------------------------------------------------------------------*)


  (* [new_definition]
   * def <name> = <formel>        modus = CLASSIC
   * axiom <name> = <formel>      modus = AXIOM
   * lemma <name> = <formel>      modus = LEMMA
   * - lege temporaeres goal aus formel an
   * - bestimme den Typ t der formel
   * - update Definitionen Hashtable keyword = name
   * - lege freie variable <name> des Typs t an
   *)
  fun new_definition ( goal, name, formel, modus) = 
      let
	  val (_ , pters, gamma, def ) = goal

	  (* lege temporaeres goal an mit formel *)
	  val (new_type_envs, def_pters, _ , _ ) =  first_goal(goal, formel)

	  val new_type_env = hd(new_type_envs)
	  val def_pter = case (hd(def_pters)) of (* the desired expression *)
	                  (IMP(C "TRUE", p)) => p
			    | _ => raise Error ("interner Fehler xy\n")

	  (* take the type of the def_pter *)
	  val def_type = ( Typecheck.get_type (def_pter) 
			                   ( (fn c => Env.typc (new_type_env, c)),
  					     (fn i => Env.typi (new_type_env, i)),
					     (Env.typmaps2b(new_type_env))        )
			 ) handle Typecheck.illtyped e => 
	      raise Error ("Definition/Axiom/Lemma was illtyped: ^ e \n")

	  (* update the Definition table *)
	  (* create a ref of current Environment 
	   * only saved if mode = LEMMA 
	   *)
	  val envref = ref (new_type_env) 
	  val newdef = Definition.insert_def def (name, def_pter, modus) envref
	     
	  (* !!! DEBUG !!! *)
	  (* val _ = print ("AXIOM " ^ name ^ " :" ^ Visual.pter2string(def_pter) ^ "\n") *)
	  (* update the goals *)
	  val newgoal' = (new_type_envs, pters, gamma, newdef)
	  (* makes definition <name> as a new free variable known *)
	  val newgoal  = update_ftypeenv(newgoal', name, def_type)
      in
	  newgoal
      end


  (* [save_goal]
   * Saves the current solved goal under AXIOM name 
   *)

  fun save_goal (goal, name, pter) = 
      let
	  val (type_envs, pters, gamma, def) = goal
	  val dummyenvref = ref (Env.empty())
	  val newdef = Definition.insert_def def (name, pter, AXIOM) dummyenvref
	  val newgoal = (type_envs, pters, gamma, newdef)
      in
	  newgoal
      end


  (* [check_def]
   * gibt true zurueck, wenn der Teilausdruck for eine Definition beinhaltet 
   * ruft Definition.check_def auf
   *)
  fun check_def(goal, for) = 
      let
	  val (type_envs,_,_, def ) = goal
	  val type_env = hd(type_envs)
      in
	  Definition.check_def def (Env.freemapi2s(type_env)) for
      end

  (* [translate_def]
   * ersetzt alle auf erster Ebene gefundenen Definitionen innerhalb
   * des Terms <pter>
   *)

  fun translate_def(goal, pter) = 
      let
	  val (type_envs,_,_,def) = goal
	  val type_env = hd(type_envs)
      in
	  Definition.translate_def def (Env.freemapi2s(type_env)) pter
      end

   (* [get_definition] 
   * gibt die fuer den Key gespeicherte Definition aus
   * der Hashtable zurueck 
   *)

  fun get_definition(goal,name) =
      let
	  val (_,_,_,def) = goal
      in
	  Definition.get_term def name
      end

 (* [get_definitions_kind] 
   * gibt die Art der fuer den Key gespeicherten Definition aus
   * der Hashtable zurueck, sprich CLASSIC | LEMMA | AXIOM
   *)

  fun get_definitions_kind(goal, name) =
      let
	  val ( _,_,_,def ) = goal
      in 
	  Definition.get_kind def name
      end
	  
  (* [change_definitions_kind]
   * aendert den Modus der Definition 
   * i.e. von LEMMA -> AXIOM ,wenn Lemma
   * bereits hinzugefuegt wurde! (vgl.Apply.sml)
   *)
  fun change_definitions_kind (goal, name, kind) =
      let
	  val ( _,_,_,def ) = goal
      in 
	  Definition.change_kind def (name,kind)
      end



  (* [get_envref]
   * gibt die Referenz auf die Umgebung zurueck, die
   * bei der Definition des Lemmas aktuell war
   *)
  fun get_envref(goal, name) =
      let
	  val ( _,_,_,def ) = goal
      in 
	  Definition.get_envref def name
      end
(* --------------------------------------------------------------------*)
(* |                     ENDE Definitionen Methoden                   |*)
(* --------------------------------------------------------------------*)

 
(* --------------------------------------------------------------------*)
(* | BEGIN Tunnelfunktionen ( Aufruf dergleichen Methoden in Env.sml )|*)
(* --------------------------------------------------------------------*)

(* Fuer Erklaerung der Methoden, siehe entsprechende Methoden 
 * innerhalb der Struktur Env.sml
 *)

  fun freemaps2i(oldgoal) = let 
			    val (envs,_,_,_) = oldgoal
			    val env = hd(envs)
			in
			    ( Env.freemaps2i(env))
			end

  fun freemapi2s(oldgoal) = let 
			    val (envs,_,_,_) = oldgoal
			    val env = hd(envs)
			in
			    ( Env.freemapi2s(env))
			end

  fun bndmapi2s (oldgoal ) = let
				 val (envs,_,_,_) = oldgoal
				 val env = hd(envs)
			     in
				 (Env.bndmapi2s (env))
			     end

  fun bndmaps2i (oldgoal ) = let
				 val (envs,_,_,_) = oldgoal
				 val env = hd(envs)
			     in
				 (Env.bndmaps2i (env))
			     end

   fun typi(oldgoal) = let
			  val (envs,_,_,_) = oldgoal
			  val env = hd(envs)
		      in 
			  fn i => (Env.typi(env,i))
		      end

  fun typc(oldgoal) = let
			  val (envs,_,_,_) = oldgoal
			  val env = hd(envs)
		      in 
			  fn c => (Env.typc(env,c))
		      end


  fun typemap(oldgoal) = let
			     val (envs,_,_,_) = oldgoal
			     val env = hd(envs)
			 in 
			     (Env.typmaps2b(env))
			 end

  fun nr_freevars(oldgoal) = let
				 val (envs,_,_,_) = oldgoal
				 val env = hd(envs)
			     in
				 (Env.nr_freevars(env))
			     end

  fun nr_bndvars(oldgoal) = let
				 val (envs,_,_,_) = oldgoal
				 val env = hd(envs)
			     in
				 (Env.nr_bndvars(env))
			     end


(* --------------------------------------------------------------------*)
(* |                        ENDE Tunnelfunktionen                     |*)
(* --------------------------------------------------------------------*)

                        

(* --------------------------------------------------------------------*)
(* |                     BEGIN Zusatzmethoden                         |*)
(* --------------------------------------------------------------------*)


(* [get_env]
 * gibt die aktuelle Umgebung env zurueck
 *)
  fun get_env(oldgoal) = 
      let 
	  val ( envs , _ ,_,_ ) = oldgoal
	  val env = hd(envs)
      in
	  env
      end
  
(* [set_env]
 * setzt die aktuelle Umgebung 
 *)
  fun set_env(oldgoal, newenv) = 
      let
	  val ( old_envs , pters , gamma,def) = oldgoal
      in
	  (newenv::tl(old_envs), pters, gamma, def)
      end
  
(* [get_nr_of_goals]
 * gibt die Anzahl der aktiven Goals/Subgoals zurueck
 *)
  fun get_nr_of_goals(oldgoal) = 
      let 
	  val (_, pters ,_,_) = oldgoal
      in 
	  List.length(pters)
      end

(* [cyclic_shift]
 * shiftet die pters und envs cyclic nach oben 
 * verwendet in Instance.sml
 *) 			    

  fun cyclic_shift(oldgoal) = 
      let
	  val (envs, pters, gamma , def ) = oldgoal
      in
	  case (envs, pters) of
	      (nil, nil )                => oldgoal
	    | ( (e::env) , (p::pter) )   => (env @ [e], pter @ [p], gamma, def)
	    | _                          => oldgoal
      end

(* [Print_goal]
 * printet die Goals
 *)
  fun print_goal (oldgoal) = 
      let 
	  val (envs, pters, _,_ ) = oldgoal
	  val nrgoals = List.length(pters)
          (* val _ = print ("> ") *)
	  fun print_it env pter n = Visual.print_goal (pter, Env.freemapi2s(env), Env.bndmapi2s(env) ) (n, nrgoals)

	  fun cyclic_print nil nil         n = ()
	    | cyclic_print (x::xs) (y::ys) n = ( print_it x y n; cyclic_print xs ys (n+1) )
	    | cyclic_print _ _             n = ()
             
      in
	  cyclic_print envs pters 1
      end

(* [latex_goal]
 * gibt latex Strings der angegebnen Term Liste zurueck
 *)
  fun latex_goal (oldgoal, pter) = 
      let 
	  val (envs, _ , _,_ ) = oldgoal
	  val env = hd(envs)
      in
	  Visual.latex_goals (pter,(Env.freemapi2s(env)), (Env.bndmapi2s(env)))
      end


(* --------------------------------------------------------------------*)
(* |                     ENDE Zusatzmethoden                          |*)
(* --------------------------------------------------------------------*)

end (* of struct *)
