(* Hilfe.sml - gibt einen Ueberblick der einzelnen Kommandos 
 * mit Anwendungshilfe 
 *)

fun line(n:int)= implode(List.tabulate(n,fn x=> #"-" )) ^ "\n"

val list = [ ("type X\t\t\t",      "declares a new type X\n") ,
	     ("var x:t\t\t\t",     "declares a free variable x with Type t\n"),
	     ("const x:t\t\t",     "declares a constant x with Type t\n"),
	     ("def name = TER\t\t","defines a shorthand name for TERM\n"),
	     ("lemma name = FOR\t","specifies a new lemma name, which must be proven after usage\n"),
	     ("axiom name = FOR\t","specifies a new axiom name\n"),
	     ("goal FOR\t\t",      "specifies a new theorem to proof\n"),
	     ("",line(80)),
	     ("ref\t\t\t",         "Reflexivity Rule \n"),
	     ("sym\t\t\t",         "Symmetric Rule \n"),
	     ("trans TER\t\t",    "Transitivity Rule \n"),
	     ("subst var = TER \t", "Substitution Rule \n"),
	     ("pos [nr]\t\t",      "Calls a java program to determine subterm positions \n"),
	     ("rew (pos) with beta\t","applies the Beta-Rule on position pos \n"),
	     ("rew (pos) with eta\t", "applies the Eta-Rule on position pos \n"),
	     ("rew (pos) with (N=M)\t","Rewriting Rule: introduces the new lemma (N=M) and tries \n"),
	     ("\t\t\t",             "to rewrite the subterm on position pos with the new lemma\n"),
	     ("rew (pos) with \t\t", "Rewriting Rule: rewrites the subterm on position pos\n"),
	     ("    (name/nr) (l/r)\t", "with the (l)eft/(r)ighthandside of an axiom/lemma \n"),
	     ("\t\t\t",            "(name) or a premise (nr) \n"),
	     ("",line(80)),
	     ("intro\t\t\t",       " |- x => y\t\t << intro >>\t\tx |- y\n"),
	     ("\t\t\t",            " |- x & y\t\t << intro >>\t\t|- x   (2/1) \n\t\t\t\t\t\t\t\t\t|- y   (2/2)\n"),
	     ("\t\t\t",            " |- ~x\t\t\t << intro >>\t\tx |- FALSE\n"),
	     ("\t\t\t",            " |- /\\ x. f(x)\t\t << intro >>\t\t|- f(? 0) (Gamma Var)\n"),
	     ("\t\t\t",            " |- \\/ x. f(x)\t\t << intro >>\t\t|- x\n"),
	     ("","\n"),
	     ("elim \t\t\t",       " x + y |- z\t\t << elim  >>\t\tx |- z   (2/1) \n\t\t\t\t\t\t\t\t\ty |- z   (2/2)\n"),
	     ("\t\t\t",            " \\/ x. f(x) |- z\t << elim  >>\t\tf(? 0) |- z\n"),
	     ("\t\t\t",            " /\\ x. f(x) |- z\t << elim  >>\t\tx |- z\n"),
	     ("","\n"),
	     ("instance ganr = FOR\t", " |- (? ganr) \t\t << instance >> \t |- FOR \n"),
	     ("","\n"),
	     ("cut FOR \t\t",      " x |- y\t\t\t << cut FOR >> \t\tx|-FOR + y   (1/2) \n\t\t\t\t\t\t\t\t\tx & FOR|- y   (2/2)\n"),
	     ("","\n"),
	     ("move l\t\t\t",      " [1] x \n"),
	     ("\t\t\t",            " [2] y |- z\t\t << move l >>\t\t[1] y |- ~x + z\n"),
	     ("","\n"),
	     ("move r\t\t\t",      " x |- [1] y \n"),
	     ("\t\t\t",            "      [2] z \t\t << move r >>\t\t ~y & x |- z\n"),
	     ("","\n"),
	     ("remove [l/r] (nr)\t", " removes the the premise(l) / conclusion(r) nr\n"),
	     ("shift (l/r)(u/d)[nr]\t", " Cyclic (u)p/(d)own shift of premises(l)/conclusions(r)\n"),
	     ("\t\t\t",                 " by nr positions\n"),
	     ("lam\t\t\t",         " forces a beta- and eta reduction on the current formula\n"),
	     ("simplify\t\t",      " applies the BOOL-Rule in order to simplify the formula\n"),
	     ("",line(80)),
	     ("print\t\t\t",       " outputs the current goals\n"),
	     ("labels\t\t\t",      " switches label printing mode \n"),
	     ("latex\t\t\t",       " outputs the proof to a latex file\n"),
	     ("undo\t\t\t" ,       " makes the last command undone\n"),
	     ("reset\t\t\t",       " all goals/type/var declarations will be deleted \n"),
	     ("exit\t\t\t" ,       " exits the program \n")	    
]
	   

fun main() =let 
		val _ = print ("Available Commands : \n"^line(80))
		val f = foldl ( fn ((cmd,desc), n) => print (cmd ^ desc) ) () list
	    in
		f
	    end
