(* Instance.sml implementiert die Instanziierung von Gammavariablen
 * mit Subtermen:
 * instance <i> = <expr>    substituiert innerhalb des aktiven Goals 
 *                          alle Vorkommen der Gammavariablen i durch
 *                          <expr>
 *
 * Gamma Variablen sind genauso wie Deltavaribalen innerhalb eines
 * Graphens gespeichert, bei jeder Instanziierung entscheided ein Zyklustest
 * ueber die Gueltigkeit der Substituierung
 *) 

local open Data Goal Env in


val var_order = fn (GAMMA(_,i) , GAMMA(_,j)) => if (i < j) then LESS else if ( i > j ) then GREATER else EQUAL
                  |(GAMMA(_,i) , V (j) )     => LESS (* Gammas immer kleiner als V *)
                  |(V(i) , GAMMA(j) )        => GREATER 
                  |(V(i) , V(j))             => if (i < j) then LESS else if ( i > j ) then GREATER else EQUAL
                  | (_,_) => LESS
     
(* remove_dbl loescht das Doppelauftreten der Variablen in der Liste *)
fun remove_dbl xs = foldr ( fn (x,nil) => [x] 
                               | (x,ys) => if x = hd(ys) then ys else x::ys ) 
                      nil (Listsort.sort var_order xs)

(* Checkt mittels Typechecker, ob Term <pter> den Typ <typ> besitzt *)
fun checktype (pter, typ) goal = 
    let 
	val cmap = Goal.typc(goal)
	val fmap = Goal.typi(goal)  (* Gets a int->Typ mapping function *)
	val tmap = Goal.typemap(goal)  (* Gets a string->bool mapping function *)
    in (( Typecheck.get_type pter (cmap,fmap,tmap)) = typ ) 
	handle Typecheck.illtyped _ => false
    end
			     
(* Ersetzt Gamma Variable i durch Term n, nur dann, wenn Typen uebereinstimmen *)
fun gammasub (g as GAMMA (t,i) ) n nr tycheck = 
    if (i=nr) then 
	if (tycheck t) 
	then n 
	else raise Error("Gamma Var. "^Int.toString(i)^" and Instanciation hava different Types\n")
    else g
  | gammasub (A(m,m'))           n nr tycheck = A(gammasub m n nr tycheck, gammasub m' n nr tycheck)
  | gammasub (L((t,i),m))        n nr tycheck = L((t,i), gammasub m (ETT.up 0 n 1) nr tycheck)
  | gammasub (EX((t,i),m))       n nr tycheck = EX((t,i), gammasub m (ETT.up 0 n 1) nr tycheck)
  | gammasub (ALL((t,i),m))      n nr tycheck = ALL((t,i), gammasub m (ETT.up 0 n 1) nr tycheck)
  | gammasub (AND(m,m'))         n nr tycheck = AND( gammasub m n nr tycheck, gammasub m' n nr tycheck)
  | gammasub (OR(m,m'))          n nr tycheck = OR( gammasub m n nr tycheck , gammasub m' n nr tycheck )
  | gammasub (IMP(m,m'))         n nr tycheck = IMP( gammasub m n nr tycheck , gammasub m' n nr tycheck )
  | gammasub (EQ(m,m'))          n nr tycheck = EQ( gammasub m n nr tycheck, gammasub m' n nr tycheck )
  | gammasub (NOT(m))            n nr tycheck = NOT( gammasub m n nr tycheck)
  | gammasub a                   n nr tycheck = a  (* Var und Const bleiben erhalten *)



fun subst_gamma'(goal, _,_) 0 = goal
  | subst_gamma'(oldgoal, gammanr, f) n =
 let
    val pter = Goal.get_first_goal(oldgoal)
    val instgoal = Goal.first_goal( Goal.init_goal_env(oldgoal), f )

    val inst = 
	case (Goal.get_first_goal (instgoal)) of
	    IMP(_,t) => t
	  | _ => raise Error ("NEVER HAPPEND\n")
	
    val freevarlist = remove_dbl(ETT.getfreevars(inst))
   (* DEBUG  val _ = List.app( fn a => print (Visual.pter2string a)) freevarlist *)
	
    val typcheck = fn t => checktype (inst, t) oldgoal

    val mergegoal = Goal.merge_envs(oldgoal,instgoal)
   (* wenn es erlaubt ist zu substituieren, dann tue es ;) *)
    val newgoal = if ( Goal.substitute_gamma oldgoal (gammanr, freevarlist) )
		  then 
		      let 
			  val tempgoal = Goal.update_first_goal(mergegoal, gammasub pter inst gammanr typcheck)
			  (* Ueberpruefe, ob freie Variablen gekapert wurden *)
			  val kaperfreegoal = Goal.kaper_check(tempgoal)
		      in
			  kaperfreegoal
		      end
		  else raise Error ("There are free variables in instanciation, not known at gamma creation\n")
in
    subst_gamma'(Goal.cyclic_shift(newgoal), gammanr, f) (n-1)
end



(* Hauptroutine zum Substituieren *)
(* ruft fuer jedes Subgoal subst_gamma' auf *)

fun subst_gamma(oldgoal, gammanr, f) = 
let
    val n = Goal.get_nr_of_goals(oldgoal)
in
    subst_gamma'(oldgoal, gammanr, f) n
end

	 

fun main(goal,i,formel) = let
			      val newgoal = subst_gamma(goal,i,formel)
			  in
			      (Goal.print_goal(newgoal) ; newgoal )
			  end

end (* of local open *)