(* Typechecker : pter -> (int -> ty) -> (string -> bool) ->  bool *)
(*               term     freevar map      typ map           result  *)

local open Data in

exception illtyped of string
fun cons t (g : var -> ty) = fn x => if (x=0) then t else g (x-1)



(* [checktypetree]
 * checktypetree : ty -> ( string -> bool ) -> bool
 * checkt, ob alle Blaetter im Typbaum ty bekannt sind
 *)
fun checktypetree (TA(a,b)) map = (checktypetree a map) andalso (checktypetree b map)
  | checktypetree (TC t) map = map t

(* basetype_test : (string->bool) -> ( ty -> bool )  
 * gibt ty -> bool funktion zurueck, die checkt, ob alle
 * Blaetter Typen in ty bekannt sind!
 *)
val basetype_test  = fn (map: string -> bool) => fn (t:ty) => checktypetree t map

(* Gibt die Ordnung eines Typs an *)
fun type_ordnung (TC _ )    =  0
  | type_ordnung (TA (a,b)) =  Int.max(1 + (type_ordnung a), type_ordnung b)


fun tycheck' (C c) tyc _ _ = tyc c
  | tycheck' (V x) _ ga _  = ga x
  | tycheck' (GAMMA (t,_)) _ _ _ = t
  | tycheck' (A(m,m')) tyc ga ta = 
    ( case (tycheck' m tyc ga ta) of
	  (TA (t1, t2)) => if (t1 = (tycheck' m' tyc ga ta) ) then t2 
			   else raise illtyped ("Ill typed Application")
	| _             => raise illtyped ("Ill typed Application")
	      )

  | tycheck' (L((t,_),e)) tyc ga ta = 
	  if ( ta t ) then TA(t, tycheck' e tyc (cons t ga) ta)
	  else raise illtyped ("Ill typed Lambda construct")

  | tycheck' (EX((t,_),e)) tyc ga ta = 
	      if ( ta t ) then 
		  ( case tycheck' e tyc (cons t ga) ta of
			(t' as (TC ("bool"))) => t'
		      | _ => raise illtyped "Ill typed Existence Quantifier" )
	      else raise illtyped ("Ill typed - Unknown base types in Ex. Quantifier")
				  
  | tycheck' (ALL((t,_),e)) tyc ga ta = 
		  if (ta t) then 
		      ( case tycheck' e tyc (cons t ga) ta of
			    (t' as (TC ("bool"))) => t'
			  | _ => raise illtyped "Ill typed All Quantifier" )
		  else raise illtyped ("Ill typed - Unknown base types in All Quantifier") 
				  
  | tycheck' (AND(m,m')) tyc ga ta = 
		      (case ( tycheck' m tyc ga ta , tycheck' m' tyc ga ta ) of
			   (TC "bool" , TC "bool" ) => TC "bool"
			 | _ => raise illtyped ("Ill typed Conjunction") )

  | tycheck' (OR(m,m')) tyc ga ta = 
			   (case ( tycheck' m tyc ga ta , tycheck' m' tyc ga ta ) of
				(TC "bool" , TC "bool" ) => TC "bool"
			      | _ => raise illtyped ("Ill typed Disjunction" ) )

  | tycheck' (IMP(m,m')) tyc ga ta =
				(case ( tycheck' m tyc ga ta , tycheck' m' tyc ga ta ) of
				     (TC "bool" , TC "bool" ) => TC "bool"
				   | _ => raise illtyped ("Ill typed Implication") )

  | tycheck' (EQ(m,m')) tyc ga ta = 
	   let val t1 = tycheck' m tyc ga ta
	       val t2 = tycheck' m' tyc ga ta
	   in if ( t1=t2 ) 
		  then TC "bool"
	      else raise illtyped ("Ill Typed Equation")
	   end

  | tycheck' (NOT(m)) tyc ga ta = 
	   ( case ( tycheck' m tyc ga ta) of
		 (t' as (TC "bool"))  => t'
	       | _ => raise illtyped ("Ill typed NOT Application") )
		 
fun tycheck pt (tyc, ga, ta) = (tycheck' pt tyc ga (basetype_test ta) ; true ) handle illtyped _ => false

fun get_type pt (tyc, ga, ta) = tycheck' pt tyc ga (basetype_test ta) (* returns the type of an pter *)

end (* of local open *)