structure Unify :> Unify =
struct 

    open Data

    (* Der mgu(t1,t2) ist eine Liste von Tupeln,
     * wobei die erste Komponente Variablen enthalten,
     * die zweite Terme 
     *)

    exception notunifiable
    type unifier = (int * pter) list


    fun union nil ys = ys
     |  union (x::xs) ys = if not(List.exists( fn t => t = x ) ys ) 
			       then union xs (x::ys) 
			       else union xs ys

    fun sub b ts = foldl ( fn ((i,t), newb ) => ( ETT.sub newb i t ) ) b ts

    (* ------------------------------------------------- *)
    (* |            BEGIN Hauptmethoden                | *)
    (* ------------------------------------------------- *)

    fun show_unifier uni = 
	foldl ( fn ((i, n),_)=> print (" (V "^Int.toString(i)^") -> (" ^ Visual.pter2string(n) ^ ")\n") ) () uni


    fun getUnifier' (A(a,b)) (A(c,d)) us t = 
	let 
	    val ts = getUnifier' a c us t
	 (* val _ = print ("------------------ DEBUG ------------------\n")
	    val _ = print ("A(a,b)   b = " ^ Visual.pter2string(b) ^ "\n")
            val _ = print ("A(c,d)   d = " ^ Visual.pter2string(d) ^ "\n")
	    val _ = print ("TEMP UNIFIER : \n")
	    val _ = show_unifier ts                                             *)
(*	    val substb = sub b ts               *)
	    val substd = sub d ts 
	 (* val _ = print ("A(a,b)   b = " ^ Visual.pter2string(substb) ^ "\n")
            val _ = print ("A(c,d)   d = " ^ Visual.pter2string(substd) ^ "\n")
            val _ = print ("---------------- DEBUG ENDE ---------------\n")     *)
	in getUnifier' b substd (union ts us ) t
	end


      | getUnifier' (AND(a,b)) (AND(c,d)) us t = 
	let 
	    val ts = getUnifier' a c us t
(*	    val substb = sub b ts          *)
	    val substd = sub d ts
	in getUnifier' b substd (union ts us ) t
	end
    
      | getUnifier' (OR(a,b)) (OR(c,d)) us t = 
	let 
	    val ts = getUnifier' a c us t
(*	    val substb = sub b ts          *)
	    val substd = sub d ts
	in getUnifier' b substd (union ts us ) t
	end

      | getUnifier' (IMP(a,b)) (IMP(c,d)) us t = 
	let 
	    val ts = getUnifier' a c us t
(*	    val substb = sub b ts          *)
	    val substd = sub d ts
	in getUnifier' b substd (union ts us ) t
	end

      | getUnifier' (EQ(a,b)) (EQ(c,d)) us t = 
	let 
	    val ts = getUnifier' a c us t
(*	    val substb = sub b ts          *)
	    val substd = sub d ts
	in getUnifier' b substd (union ts us ) t
	end

      | getUnifier' (NOT(a)) (NOT(b)) us t= 
	getUnifier' a b us t

      | getUnifier' (ALL((t1,i1),m)) (ALL((t2,i2),n)) us t =
	let
	    (* Pruefe, ob die Typen der Quantoren Basistypen sind und uebereinstimmen *)
	    val _ = if ((t1 = t2)) then ()  (* andalso ((Typecheck.type_ordnung t1) = 0) ) then () *)
		    else raise notunifiable
	in
	    getUnifier' (ETT.down 1 m 1) (ETT.down 1 n 1) us t
	end
		
      | getUnifier' (EX((t1,i1),m)) (EX((t2,i2),n)) us t =
	let
	    (* Pruefe, ob die Typen der Quantoren Basistypen sind und uebereinstimmen *)
	    val _ = if ((t1 = t2)) then () (* andalso ((Typecheck.type_ordnung t1) = 0) ) then () *)
		    else raise notunifiable
	in
	    getUnifier' (ETT.down 1 m 1) (ETT.down 1 n 1) us t
	end

      | getUnifier' (L((t1,i1),m)) (L((t2,i2),n)) us t = 
	let
	    (* Pruefe, ob die Typen der Quantoren Basistypen sind und uebereinstimmen *)
	    val _ = if ((t1 = t2)) then ()  (* andalso ((Typecheck.type_ordnung t1) = 0) ) then () *) 
		    else raise notunifiable
	in
	    getUnifier' (ETT.down 1 m 1) (ETT.down 1 n 1) us t
	end

      | getUnifier'  (C s) (C t) us _ = if (s=t) then us else
	                              raise notunifiable

      | getUnifier'  (V i) (V j) us _ = if (i <> j) then ( union [(j, V i)] us ) else us

      | getUnifier'  n     (V i) us t = if (i >= t) then
					    if not(ETT.free i n) then ( union [(i, n)] us )
					    else raise notunifiable
				      else raise notunifiable

      (* Normalerweise sollte der naechste Fall unifizieren, aber
       * Vereinbarung : Unifiziere nur den freeterm ( Axiom / Lemma ) und nicht
       *                den fixen Goal Subterm !! 
       *)
   (*   | getUnifier'  (V i) n     us = if not(ETT.free i n) then ( union [(i, n)] us )
				      else raise notunifiable  *)

      | getUnifier' _ _ _ _ = raise notunifiable



    fun getUnifier (fixterm, freeterm) threshold  = 
	( let
	      val unifier  = ( getUnifier' fixterm freeterm nil threshold )
	  in
	      unifier
	  end
	 )
	      handle notunifiable => 
		  raise Error ("Axiom and Subterm not FO-Unifiable \n")
		      

    fun subst' freeterm unifs 0 = freeterm
      | subst' freeterm unifs n = subst' (sub freeterm unifs) unifs (n-1)

    fun subst_wrt_unifier (freeterm, unifs ) =
	(* Muss die Substitutionsliste n mal durchlaufen lassen :
	 *  Worst case : 
	 *                   [ V 0  -->  C 0 ,
	 *		       V 1  -->  V 0 ,
	 *		       V 2  -->  V 1 ] 
	 * ==>    Bei einmaliger Substitution : Aus V 2 -> V 1 ( genuegt nicht ) 
	 *)
	subst' freeterm unifs (List.length(unifs))

end (* of struct *)



